#include <stdio.h>
#include "ports.h"
#include "cprintf.h"

// The API for the test case is as follows:
//
// R2 points to a ROM buffer of 100 bytes
// R3 points to a RAM buffer of 100 bytes
// the C flag is cleared
// Don't change R0 or R1 or R4 - the others are all OK to change.

typedef int (*RAMFUNC)(volatile uint16_t *, const char *, char *, int, int);

typedef struct {
  RAMFUNC test;
  const char *desc;
} Tests;

extern Tests tests[];
extern int n_tests;
extern int reps;

static const char buf_a[100];
static char buf_b[100];
static char buffer[20000];
static int bufp, bufs, bufe;

#define FUNC() bfunc(&(TPU[2].tcnt), buf_a, buf_b, 0, 0x3f800000)
#define TFUNC(i) tests[i].test(&(TPU[2].tcnt), buf_a, buf_b, 0, 0x3f800000)

#define S(x) buffer[bufp++] = x

#define REPS 100

int need_end;

static void
stuff_begin ()
{
  // clrpsw I
  S(0x7f); S(0xb8);
  // mov.w [r1],r4
  S(0xdc); S(0x14);
  bufs = bufp;
  need_end = 1;
}

static void
stuff_end ()
{
  int i, j;
  bufe = bufp;
  for (i=0; i<REPS-1; i++)
    {
      for (j=bufs; j<bufe; j++)
	S (buffer[j]);
    }
  // mov.w [r1],r1
  S(0xdc); S(0x11);
  // setpsw I
  S(0x7f); S(0xa8);
  // sub r4, r1
  S(0x43); S(0x41);
  // mov.w r1, r1
  S(0xdf); S(0x11);
  need_end = 0;
}

static void
stuff_prologue (int prompt)
{
  bufp = 0;
  if (prompt)
    tty_puts ("\n*ASM\n");
  // pushm r2-r14
  S(0x6e); S(0x2e);
  // fdiv r5,r5 - 16 clocks - fill the prefetch queue
  S(0xfc); S(0x93); S(0x55);
  // clrpsw C
  S(0x7f); S(0xb0);
  stuff_begin ();
}

static void
stuff_epilogue ()
{
  if (need_end)
    stuff_end ();
  // popm r2-r14
  S(0x6f); S(0x2e);
  // rts
  S(0x02);
}

main()
{
  int c, b, got_one;
  RAMFUNC bfunc;
  int cycles, base_cycles;

  /* Power up the peripherals.  */
  MSTPCRA = 0;
  MSTPCRB = 0;
  MSTPCRC = 0;

  DDR[10] = 0xff;
  DR[10] = 0x01;

  /* Timer */
  TPU[2].tcr = 0x00;	/* TPU2 counts at 50 MHz */
  TSTRA |= 0x04;	/* start timers */

  set_system_clock_speed ();
  asm("setpsw i");

  tty_init ();
  tty_puts("DJ Was Here\r\n");

  DR[10] = 0x02;

  b = 0;
  bfunc = (RAMFUNC)&buffer;

  stuff_prologue(0);
  stuff_begin();
  S(0x03);
  stuff_end();
  stuff_epilogue();
  base_cycles = FUNC();
  cprintf("base cycles %d\n", base_cycles);
  base_cycles -= REPS;

  base_cycles = -1;

  for (c=1; c<n_tests; c++)
    {
      if (tests[c].test)
	{
	  cycles = TFUNC(c);
	  if (base_cycles == -1)
	    base_cycles = cycles - reps;

	  if (cycles % 50 < 5)
	    cycles -= cycles % 25;

	  cprintf("%d\t%d\t%d\t", (int)((cycles - base_cycles + reps/2) / reps), cycles, base_cycles);
	}
      else
	cprintf("\t\t\t");

      tty_puts (tests[c].desc);
    }

  tty_putc(3);
  tty_putc(3);
  tty_putc(0);
  tty_putc(0);

  stuff_prologue (1);
  got_one = 0;

  while (1)
    {

      c = tty_getc ();
      tty_putc (c);

      if (c >= '0' && c <= '9')
	{
	  b = b * 16 + c - '0';
	  got_one = 1;
	}

      else if (c >= 'a' && c <= 'f')
	{
	  b = b * 16 + c - 'a' + 10;
	  got_one = 1;
	}

      else if (c >= 'A' && c <= 'F')
	{
	  b = b * 16 + c - 'A' + 10;
	  got_one = 1;
	}

      else if (got_one)
	{
	  got_one = 0;
	  buffer[bufp++] = b;
	  b = 0;
	}

      if (c == '<')
	stuff_begin ();

      if (c == '>')
	stuff_end ();

      if (c == '\n' || c == '\r')
	{
	  int i;

	  stuff_epilogue ();

#if 0
	  tty_puts("\nBuffer:");
	  for (i=0; i<bufp; i++)
	    cprintf(" %b", buffer[i]);
	  tty_putc('\n');
#endif

	  DR[10] = 1;
	  cycles = FUNC();
	  DR[10] = 2;
	  tty_puts("\n*CLOCK ");
	  cprintf("%d %d %d\n", (int)((cycles - base_cycles + REPS/2) / REPS), cycles, base_cycles);
	  DR[10] = 3;
	  stuff_prologue (1);
	}

    }
}
