#include "iodefine.h"
#include "sdram.h"

#define BGA176 1

void
sdram_init (void)
{
  int i;

  SYSTEM.SYSCR0.WORD = 0x5A03;   /* enable the external bus...must also set key and internal ROME bit */
  IOPORT.PF3BUS.BYTE = 0x00;     /* Disable A16-A24 */
  IOPORT.PF4BUS.BYTE = 0xFF;     /* Enable A0-A15 */
  IOPORT.PF5BUS.BYTE = 0x10;     /* Enable D8-D15 */
  IOPORT.PF6BUS.BYTE = 0xD0;     /* Enable SDRAM pins */
  
  { int delay = 100*100; while (--delay > 0); } /* delay at least 100uS after SDCLK active */

  /* SDRAM enabled */
  IOPORT.PF6BUS.BYTE = 0xd0;
  /* 16-bit SDRAM, disabled - this gives us the 10 clocks we need to stabilize.  */
  BSC.SDCCR.BYTE = 0x00;

  for (i=10; i; i--)
    asm("nop");

  /* At least 10 clocks must pass since we enabled the SDRAM pins above.  */
  /* Continuous read enabled */
  BSC.SDAMOD.BYTE = 1;
  /* precharge 3 cycles, 15 refresh 0 cycles */
  BSC.SDIR.WORD = 0x00f0;
  /* Initialize */
  BSC.SDICR.BYTE = 0x01;
  while (BSC.SDSR.BYTE)
    ;
  /* SDRAM mode register */
  BSC.SDMOD.WORD = 0x0220;
  while (BSC.SDSR.BYTE)
    ;
  /* Same endian as MCU */
  BSC.SDCMOD.BYTE = 0x00;
  /* CL = 2, WR = 2, RP = 1, RCD = 1, RAS = 3 */
  BSC.SDTR.LONG = 0x00020102;
  /* 9-bit address shift */
  BSC.SDADR.BYTE = 1;
  /* Auto-refresh: 64 ms / 20 ns */
  BSC.SDRFCR.WORD = 3200 | 0xf000;
  BSC.SDRFEN.BYTE = 1;
  /* 16-bit, enabled */
  BSC.SDCCR.BYTE = 0x01;

  /* We use CS3-B.  */
  IOPORT.PF0CSE.BIT.CS3E = 1;
  IOPORT.PF2CSS.BIT.CS3S = 1;

  /* CS3: native endian, 16-bit, enabled.  */
  BSC.CS3CR.WORD = 0x0001;
  BSC.CS3WCR2.BIT.CSON = 0;

  BSC.CS3WCR2.BIT.RDON = 1;
  BSC.CS3WCR1.BIT.CSRWAIT = 2;
  BSC.CS3WCR2.BIT.CSROFF = 1;

  BSC.CS3WCR2.BIT.WDON = 0;
  BSC.CS3WCR2.BIT.WRON = 0;
  BSC.CS3WCR1.BIT.CSWWAIT = 0;
  BSC.CS3WCR2.BIT.WDOFF = 1;
  BSC.CS3WCR2.BIT.CSWOFF = 1;

  BSC.CS3REC.BIT.RRCV = 0;
  BSC.CS3REC.BIT.WRCV = 0;

}

void *
sdram_base (void)
{
  return (void *) 0x08000000;
}

static unsigned long saved_sdram_size = 0;

unsigned long
sdram_size (void)
{
  volatile unsigned char *sdram = (unsigned char *) sdram_base ();
  volatile unsigned long *sdram_l;
  unsigned long i;

  if (saved_sdram_size == 0)
    {
      for (i = 128 * 1024 * 1024; i > 4; i >>= 1)
	{
	  sdram_l =  (volatile unsigned long *) (sdram + (i & 0x7ffffff));
	  *sdram_l = i;
	}
      sdram_l =  (volatile unsigned long *) sdram;
      saved_sdram_size = *sdram_l;
    }
    return saved_sdram_size;
}
